@@ -1,5 +1,5 @@
 /* GENIUS Calculator
- * Copyright (C) 1997-2007 Jiri (George) Lebl
+ * Copyright (C) 1997-2009 Jiri (George) Lebl
  *
  * Author: Jiri (George) Lebl
  *
@@ -34,31 +34,18 @@
 
 /*the context stack structure*/
 typedef struct _GelDictContext {
-	GSList *stack;
-	GSList *subststack;
-	GSList *stackname;
+	GelContextFrame *stack;
+	GelContextFrame *global_frame;
 	int top;
 } GelDictContext;
 
-static GelDictContext context = {NULL, NULL, NULL, -1};
+static GelDictContext context = {NULL, NULL, -1};
 
 static GHashTable *dictionary;
 
-extern GHashTable *uncompiled;
-
 extern const char *genius_toplevels[];
 extern const char *genius_operators[];
 
-GelEFunc *free_funcs = NULL;
-
-#define GET_NEW_FUNC(n) \
-	if (free_funcs == NULL) {			\
-		n = g_new (GelEFunc, 1);		\
-	} else {					\
-		n = free_funcs;				\
-		free_funcs = free_funcs->data.next;	\
-	}
-
 /*return current context number (0 is global, -1 is uninitialized)*/
 int
 d_curcontext (void)
@@ -72,8 +59,7 @@ d_makebifunc (GelToken *id, GelBIFunction f, int nargs)
 {
 	GelEFunc *n;
 
-	GET_NEW_FUNC (n);
-	memset (n, 0, sizeof (GelEFunc));
+	n = g_slice_new0 (GelEFunc);
 	n->id = id;
 	n->data.func = f;
 	n->nargs = nargs;
@@ -83,6 +69,7 @@ d_makebifunc (GelToken *id, GelBIFunction f, int nargs)
 	n->vararg = 0;
 	n->on_subst_list = 0;
 	n->named_args = NULL;
+	n->local_idents = NULL;
 	n->extra_dict = NULL;
 	n->no_mod_all_args = FALSE;
 	n->propagate_mod = FALSE;
@@ -98,8 +85,7 @@ d_makeufunc (GelToken *id, GelETree *value, GSList *argnames, int nargs,
 {
 	GelEFunc *n;
 
-	GET_NEW_FUNC (n);
-	memset (n, 0, sizeof (GelEFunc));
+	n = g_slice_new0 (GelEFunc);
 
 	n->id = id;
 	n->data.user = value;
@@ -108,12 +94,14 @@ d_makeufunc (GelToken *id, GelETree *value, GSList *argnames, int nargs,
 	n->context = context.top;
 	n->type = GEL_USER_FUNC;
 
-	/* look at the memset
+	/* look at the new0
 	n->vararg = 0;
 	n->on_subst_list = 0;
 	n->no_mod_all_args = FALSE;
 	n->propagate_mod = FALSE;
 	n->extra_dict = NULL;
+	n->local_all = 0;
+	n->local_idents = NULL;
 	*/
 
 	if (extra_dict != NULL) {
@@ -132,15 +120,14 @@ d_makevfunc(GelToken *id, GelETree *value)
 {
 	GelEFunc *n;
 
-	GET_NEW_FUNC (n);
-	memset (n, 0, sizeof (GelEFunc));
+	n = g_slice_new0 (GelEFunc);
 
 	n->id = id;
 	n->data.user = value;
 	n->context = context.top;
 	n->type = GEL_VARIABLE_FUNC;
 
-	/* look at the memset
+	/* look at the new0
 	n->nargs = 0;
 	n->vararg = 0;
 	n->on_subst_list = 0;
@@ -148,6 +135,8 @@ d_makevfunc(GelToken *id, GelETree *value)
 	n->extra_dict = NULL;
 	n->no_mod_all_args = FALSE;
 	n->propagate_mod = FALSE;
+	n->local_all = 0;
+	n->local_idents = NULL;
 	*/
 
 	return n;
@@ -159,19 +148,19 @@ d_makereffunc(GelToken *id, GelEFunc *ref)
 {
 	GelEFunc *n;
 
-	GET_NEW_FUNC (n);
-	memset (n, 0, sizeof (GelEFunc));
+	n = g_slice_new0 (GelEFunc);
 
 	n->id = id;
 	n->data.ref = ref;
 	n->context = context.top;
 	n->type = GEL_REFERENCE_FUNC;
 
-	/* look at the memset
+	/* look at the new0
 	n->nargs = 0;
 	n->vararg = 0;
 	n->on_subst_list = 0;
 	n->named_args = NULL;
+	n->local_idents = NULL;
 	n->extra_dict = NULL;
 	n->no_mod_all_args = FALSE;
 	n->propagate_mod = FALSE;
@@ -187,15 +176,16 @@ d_copyfunc(GelEFunc *o)
 	GelEFunc *n;
 	GSList *li;
 
-	GET_NEW_FUNC (n);
-	memcpy (n, o, sizeof (GelEFunc));
+	n = g_slice_dup (GelEFunc, o);
 
 	if(n->type == GEL_USER_FUNC ||
 	   n->type == GEL_VARIABLE_FUNC) {
 		D_ENSURE_USER_BODY (o);
-		n->data.user=copynode(o->data.user);
+		n->data.user = gel_copynode(o->data.user);
 	}
 	n->named_args = g_slist_copy (o->named_args);
+	n->local_idents = g_slist_copy (o->local_idents);
+	n->subst_dict = g_slist_copy (o->subst_dict);
 
 	n->extra_dict = g_slist_copy (o->extra_dict);
 	for (li = n->extra_dict; li != NULL; li = li->next)
@@ -215,8 +205,11 @@ d_makerealfunc(GelEFunc *o,GelToken *id, gboolean use)
 {
 	GelEFunc *n;
 
-	GET_NEW_FUNC (n);
-	memcpy (n, o, sizeof (GelEFunc));
+	n = g_slice_dup (GelEFunc, o);
+
+	/* never copy is_local! */
+	n->is_local = 0;
+
 	n->id = id;
 	if (o->symbolic_id == NULL)
 		n->symbolic_id = o->id;
@@ -231,21 +224,22 @@ d_makerealfunc(GelEFunc *o,GelToken *id, gboolean use)
 			n->data.user = o->data.user;
 			o->data.user = NULL;
 		} else
-			n->data.user = copynode(o->data.user);
+			n->data.user = gel_copynode(o->data.user);
 	}
 	if(use) {
-		o->named_args = NULL;
-		o->named_args = 0;
-	} else
-		n->named_args = g_slist_copy(o->named_args);
-
-	if (use) {
 		o->extra_dict = NULL;
+		o->named_args = NULL;
+		o->local_idents = NULL;
+		o->subst_dict = NULL;
+		o->nargs = 0;
 	} else {
 		GSList *li;
 		n->extra_dict = g_slist_copy (o->extra_dict);
 		for (li = n->extra_dict; li != NULL; li = li->next)
 			li->data = d_copyfunc (li->data);
+		n->named_args = g_slist_copy (o->named_args);
+		n->local_idents = g_slist_copy (o->local_idents);
+		n->subst_dict = g_slist_copy (o->subst_dict);
 	}
 
 	if (n->on_subst_list) {
@@ -277,24 +271,37 @@ d_setrealfunc(GelEFunc *n,GelEFunc *fake, gboolean use)
 			n->data.user = fake->data.user;
 			fake->data.user = NULL;
 		} else
-			n->data.user = copynode(fake->data.user);
+			n->data.user = gel_copynode(fake->data.user);
 	}
 
+	n->never_on_subst_list = fake->never_on_subst_list;
+	n->vararg = fake->vararg;
+	n->propagate_mod = fake->propagate_mod;
+	n->no_mod_all_args = fake->no_mod_all_args;
+	n->local_all = fake->local_all;
+	/* Never copy is_local */
+	n->built_subst_dict = fake->built_subst_dict;
+
 	if(use) {
 		n->named_args = fake->named_args;
+		n->local_idents = fake->local_idents;
+		n->subst_dict = fake->subst_dict;
 		n->nargs = fake->nargs;
 		fake->named_args = NULL;
+		fake->local_idents = NULL;
 		fake->nargs = 0;
-	} else {
-		n->named_args = g_slist_copy(fake->named_args);
-		n->nargs = fake->nargs;
-	}
+		fake->subst_dict = NULL;
 
-	if (use) {
 		n->extra_dict = fake->extra_dict;
 		fake->extra_dict = NULL;
 	} else {
 		GSList *li;
+
+		n->named_args = g_slist_copy (fake->named_args);
+		n->local_idents = g_slist_copy (fake->local_idents);
+		n->subst_dict = g_slist_copy (fake->subst_dict);
+		n->nargs = fake->nargs;
+
 		n->extra_dict = g_slist_copy (fake->extra_dict);
 		for (li = n->extra_dict; li != NULL; li = li->next)
 			li->data = d_copyfunc (li->data);
@@ -312,9 +319,8 @@ d_initcontext(void)
 
 	context.top = 0; /*0 means that element 0 exists!*/
 	/*add an empty dictionary*/
-	context.stack = g_slist_prepend (NULL,NULL);
-	context.subststack = g_slist_prepend (NULL,NULL);
-	context.stackname = g_slist_prepend (NULL,NULL);
+	context.stack = g_slice_new0 (GelContextFrame);
+	context.global_frame = context.stack;
 
 	dictionary = g_hash_table_new (g_str_hash, g_str_equal);
 
@@ -346,18 +352,24 @@ d_addfunc (GelEFunc *func)
 	GelEFunc *n;
 	
 	g_return_val_if_fail (func->context == context.top, func);
+
+	if (context.stack->local_all)
+		func->is_local = 1;
 	
 	/*we already found it (in current context)*/
 	n = d_lookup_local(func->id);
 	if(n) {
-		d_replacefunc(n,func);
+		int is_local = n->is_local;
+		d_replacefunc (n, func);
+		if (is_local)
+			n->is_local = 1;
 		return n;
 	}
 
-	context.stack->data = g_slist_prepend(context.stack->data,func);
+	context.stack->functions =
+		g_slist_prepend (context.stack->functions, func);
 	
 	func->id->refs = g_slist_prepend(func->id->refs,func);
-	func->id->curref = func;
 
 	return func;
 }
@@ -380,14 +392,10 @@ d_addfunc_global (GelEFunc *func)
 		return n;
 	}
 
-	last = g_slist_last (context.stack);
-	g_assert (last != NULL);
-
-	last->data = g_slist_prepend (last->data, func);
+	context.global_frame->functions =
+		g_slist_prepend (context.global_frame->functions, func);
 	
 	func->id->refs = g_slist_append (func->id->refs, func);
-	if (func->id->curref == NULL)
-		func->id->curref = func;
 
 	return func;
 }
@@ -413,16 +421,15 @@ GelEFunc *
 d_lookup_local(GelToken *id)
 {
 	GelEFunc *func;
-	
-	if(!id ||
-	   !id->refs)
+
+	if G_UNLIKELY (id == NULL || id->refs == NULL)
 		return NULL;
 	
 	/*the first one must be the lowest context*/
 	func = id->refs->data;
 	
 	/*not in currect context and we only want the currect context ones*/
-	if(func->context<context.top)
+	if (func->context < context.top)
 		return NULL;
 
 	return func;
@@ -433,20 +440,22 @@ GelEFunc *
 d_lookup_global_up1(GelToken *id)
 {
 	GelEFunc *func;
-	
-	if(!id ||
-	   !id->refs)
+	GSList *li;
+
+	if G_UNLIKELY (id == NULL || id->refs == NULL)
 		return NULL;
-	
+
 	/*the first one must be the lowest context*/
-	func = id->refs->data;
-	
-	if(func->context<context.top)
-		return func;
-	if(!id->refs->next)
-		return NULL;
-	
-	return id->refs->next->data;
+	li = id->refs;
+	do {
+		GelEFunc *f = li->data;
+		if ( ! f->is_local && f->context < context.top) {
+			return f;
+		}
+		li = li->next;
+	} while (li != NULL);
+
+	return NULL;
 }
 
 /*lookup a function in the dictionary but only in the toplevel context */
@@ -455,9 +464,8 @@ d_lookup_only_global (GelToken *id)
 {
 	GSList *li;
 	GelEFunc *func;
-	
-	if(!id ||
-	   !id->refs)
+
+	if G_UNLIKELY (id == NULL || id->refs == NULL)
 		return NULL;
 
 	li = id->refs;
@@ -473,16 +481,39 @@ d_lookup_only_global (GelToken *id)
 		return NULL;
 }
 
+/*lookup a function in the dictionary*/
+GelEFunc *
+d_lookup_global (GelToken *id)
+{
+	GelEFunc *func;
+	GSList *li;
+	
+	if G_UNLIKELY (id == NULL || id->refs == NULL)
+		return NULL;
+
+	/*the first one must be the lowest context*/
+	li = id->refs;
+	do {
+		GelEFunc *f = li->data;
+		if ( ! f->is_local || f->context == context.top) {
+			return f;
+		}
+		li = li->next;
+	} while (li != NULL);
+
+	return NULL;
+}
+
 GelToken *
 d_intern (const char *id)
 {
 	GelToken * tok;
 
-	if (id == NULL)
+	if G_UNLIKELY (id == NULL)
 		return NULL;
 
         tok = g_hash_table_lookup (dictionary, id);
-	if (tok == NULL) {
+	if G_UNLIKELY (tok == NULL) {
 		tok = g_new0 (GelToken, 1);
 		tok->token = g_strdup (id);
 		g_hash_table_insert (dictionary, tok->token, tok);
@@ -491,11 +522,68 @@ d_intern (const char *id)
 	return tok;
 }
 
+/* this may be inefficient as it also goes through global,
+ * but we don't assume we do this kind of thing often.  Only 
+ * done in d_delete and d_delete_global which are only done
+ * on Undefine and UndefineAll. */
+static void
+whack_from_all_contexts (GelEFunc *func)
+{
+	GelContextFrame *li;
+	for (li = context.stack; li != NULL; li = li->next) {
+		li->functions = g_slist_remove (li->functions, func);
+	}
+}
+
 gboolean
 d_delete(GelToken *id)
 {
-	/*FIXME: Delete function!*/
-	return FALSE;
+	GSList *li, *list;
+
+	g_return_val_if_fail (id != NULL && id->token != NULL, FALSE);
+
+	id->protected_ = 0;
+	id->parameter = 0;
+	id->built_in_parameter = 0;
+
+	list = id->refs;
+	id->refs = NULL;
+	for (li = list; li != NULL; li = li->next) {
+		GelEFunc *f = li->data;
+		f->id = NULL;
+		whack_from_all_contexts (f);
+		d_freefunc (f);
+	}
+	g_slist_free (list);
+
+	return TRUE;
+}
+
+gboolean
+d_delete_global(GelToken *id)
+{
+	GSList *list;
+	GelEFunc *f;
+
+	g_return_val_if_fail (id != NULL && id->token != NULL, FALSE);
+
+	id->protected_ = 0;
+	id->parameter = 0;
+	id->built_in_parameter = 0;
+
+	list = g_slist_last (id->refs);
+	f = list->data;
+
+	if (f->context == 0) {
+		/* inefficient, but it need not be */
+		id->refs = g_slist_remove_link (id->refs, list);
+
+		f->id = NULL;
+		whack_from_all_contexts (f);
+		d_freefunc (f);
+	}
+
+	return TRUE;
 }
 
 /*clear all context dictionaries and pop out all the contexts except
@@ -504,10 +592,10 @@ d_delete(GelToken *id)
 void
 d_singlecontext(void)
 {
-	if(context.stack==NULL)
+	if (context.stack == NULL)
 		d_initcontext();
 	else
-		while(context.top>0)
+		while (context.top > 0)
 			d_popcontext ();
 }
 
@@ -524,19 +612,19 @@ d_freedict (GSList *n)
 }
 
 static void
-whack_from_lists (GelEFunc *func)
+whack_from_subst_lists (GelEFunc *func)
 {
-	GSList *li;
-	for (li = context.subststack; li != NULL; li = li->next) {
-		GSList *fl = g_slist_find (li->data, func);
+	GelContextFrame *li;
+	for (li = context.stack; li != NULL; li = li->next) {
+		GSList *fl = g_slist_find (li->substlist, func);
 		if (fl != NULL) {
-			li->data = g_slist_delete_link (li->data, fl);
+			li->substlist = g_slist_delete_link (li->substlist, fl);
 			return;
 		}
 	}
 }
 
-/* Put on subst local var list for this current stack */
+/* Put on subst local var list for this current stack frame*/
 void
 d_put_on_subst_list (GelEFunc *func)
 {
@@ -544,10 +632,14 @@ d_put_on_subst_list (GelEFunc *func)
 		/* On a lower stackframe? */
 		/* weird but true.  So whack it and put it here,
 		 * it will get to the lower one eventually */
-		whack_from_lists (func);
+		whack_from_subst_lists (func);
 	}
-	context.subststack->data = 
-		g_slist_prepend (context.subststack->data, func);
+
+	if (func->context == -1)
+		func->context = context.top;
+
+	context.stack->substlist = 
+		g_slist_prepend (context.stack->substlist, func);
 	func->on_subst_list = 1;
 }
 
@@ -562,31 +654,30 @@ d_freefunc (GelEFunc *n)
 	*/
 
 	if (n->on_subst_list) {
-		whack_from_lists (n);
+		whack_from_subst_lists (n);
 		n->on_subst_list = 0;
 	}
 
 	/*if(n->id) {
 		n->id->refs = g_slist_remove(n->id->refs,n);
-		n->id->curref = n->id->refs?n->id->refs->data:NULL;
 	} */
 	if((n->type == GEL_USER_FUNC ||
 	    n->type == GEL_VARIABLE_FUNC) &&
 	   n->data.user)
 		gel_freetree(n->data.user);
-	g_slist_free(n->named_args);
+
 	for (li = n->extra_dict; li != NULL; li = li->next) {
 		d_freefunc (li->data);
 		li->data = NULL;
 	}
+
 	g_slist_free (n->extra_dict);
-#ifndef MEM_DEBUG_FRIENDLY
-	/*prepend to free list*/
-	n->data.next = free_funcs;
-	free_funcs = n;
-#else
-	g_free (n);
-#endif
+
+	g_slist_free (n->named_args);
+	g_slist_free (n->local_idents);
+	g_slist_free (n->subst_dict);
+
+	g_slice_free (GelEFunc, n);
 }
 
 /*replace old with stuff from new and free new,
@@ -596,19 +687,29 @@ void
 d_replacefunc(GelEFunc *old,GelEFunc *_new)
 {
 	GSList *li;
-
-	/* assert some things we don't deal with.  Should we? */
-	g_assert ( ! _new->on_subst_list);
-	g_assert ( ! old->on_subst_list);
+	gboolean put_on_subst = FALSE;
 
 	g_return_if_fail(old && _new);
 	g_return_if_fail(old->id == _new->id);
 
+	if (old->on_subst_list) {
+		whack_from_subst_lists (old);
+		old->on_subst_list = 0;
+	}
+
+	if (_new->on_subst_list) {
+		whack_from_subst_lists (_new);
+		_new->on_subst_list = 0;
+		put_on_subst = TRUE;
+	}
+
 	if(old->type == GEL_USER_FUNC ||
 	   old->type == GEL_VARIABLE_FUNC)
 		gel_freetree(old->data.user);
 
-	g_slist_free(old->named_args);
+	g_slist_free (old->named_args);
+	g_slist_free (old->local_idents);
+	g_slist_free (old->subst_dict);
 
 	for (li = old->extra_dict; li != NULL; li = li->next) {
 		d_freefunc (li->data);
@@ -618,13 +719,12 @@ d_replacefunc(GelEFunc *old,GelEFunc *_new)
 
 	memcpy(old,_new,sizeof(GelEFunc));
 
-#ifndef MEM_DEBUG_FRIENDLY
-	/*prepend to free list*/
-	_new->data.next = free_funcs;
-	free_funcs = _new;
-#else
-	g_free (_new);
-#endif
+	/* FIXME: this is inefficient */
+	if (put_on_subst) {
+		d_put_on_subst_list (old);
+	}
+
+	g_slice_free (GelEFunc, _new);
 }
 
 /*set_ref*/
@@ -637,9 +737,20 @@ d_set_ref(GelEFunc *n,GelEFunc *ref)
 	   n->type == GEL_VARIABLE_FUNC)
 		gel_freetree(n->data.user);
 	n->type = GEL_REFERENCE_FUNC;
-	g_slist_free(n->named_args);
+	g_slist_free (n->named_args);
+	g_slist_free (n->local_idents);
+	g_slist_free (n->subst_dict);
 	n->nargs = 0;
 	n->named_args = NULL;
+	n->subst_dict = NULL;
+
+	n->never_on_subst_list = 0;
+	n->vararg = 0;
+	n->propagate_mod = 0;
+	n->no_mod_all_args = 0;
+	n->local_all = 0;
+	/* don't touch is_local */
+	n->built_subst_dict = 0;
 	
 	n->data.ref = ref;
 }
@@ -654,78 +765,107 @@ d_set_value(GelEFunc *n,GelETree *value)
 	   n->type == GEL_VARIABLE_FUNC)
 		gel_freetree(n->data.user);
 	n->type = GEL_VARIABLE_FUNC;
-	g_slist_free(n->named_args);
+	g_slist_free (n->named_args);
+	g_slist_free (n->local_idents);
+	g_slist_free (n->subst_dict);
 	n->nargs = 0;
 	n->named_args = NULL;
-	
+	n->local_idents = NULL;
+	n->subst_dict = NULL;
+
+	n->never_on_subst_list = 0;
+	n->vararg = 0;
+	n->propagate_mod = 0;
+	n->no_mod_all_args = 0;
+	n->local_all = 0;
+	/* don't touch is_local */
+	n->built_subst_dict = 0;
+
 	n->data.user = value;
 }
 
-/*push a new dictionary onto the context stack*/
 gboolean
-d_addcontext(void)
+d_addcontext (GelEFunc *func)
 {
-	context.stack = g_slist_prepend (context.stack, NULL);
-	context.subststack = g_slist_prepend (context.subststack, NULL);
-	context.stackname = g_slist_prepend (context.stackname, NULL);
-	context.top++;
-	return TRUE;
-}
+	GelContextFrame *old = context.stack;
+	context.stack = g_slice_new0 (GelContextFrame);
+	context.stack->next = old;
+
+	if (func != NULL) {
+		context.stack->name = func->id;
+		context.stack->local_all = func->local_all;
+	}
 
-gboolean
-d_addcontext_named (GelToken *name)
-{
-	context.stack = g_slist_prepend (context.stack, NULL);
-	context.subststack = g_slist_prepend (context.subststack, NULL);
-	context.stackname = g_slist_prepend (context.stackname, name);
 	context.top++;
+
 	return TRUE;
 }
 
 /*pop the context stack*/
 void
-d_popcontext(void)
+d_popcontext (void)
 {
 	if (context.top != -1) {
 		GSList *li;
-		GSList *dict = context.stack->data;
-		GSList *subst = context.subststack->data;
+		GSList *dict = context.stack->functions;
+		GSList *subst = context.stack->substlist;
 		GSList *substlast = NULL;
+		GelContextFrame *of;
 
 		if (context.top != 0) {
-			for (li = subst; li != NULL; li = li->next) {
+			li = subst;
+			while (li != NULL) {
 				GelEFunc *func = li->data;
 				if (context.top == 1)
 					func->on_subst_list = 0;
 
-				if (func->type == GEL_USER_FUNC ||
-				    func->type == GEL_VARIABLE_FUNC) {
-					D_ENSURE_USER_BODY (func);
+				if ((func->type == GEL_USER_FUNC ||
+				     func->type == GEL_VARIABLE_FUNC) &&
+				    func->context >= context.top) {
+					D_ENSURE_SUBST_DICT (func);
 					func->extra_dict =
-						gel_subst_local_vars (func->extra_dict,
-								      func->data.user);
+						gel_subst_local_vars (func->extra_dict, &(func->subst_dict));
+					/* With substitution, context of the function is
+					   lowered */
+					if (func->context >= context.top)
+						func->context = context.top - 1;
+
+					/* If subst_dict goes to NULL, we no
+					 * longer need to keep this around on
+					 * the subst_lists */
+					if (func->subst_dict == NULL) {
+						if (substlast != NULL) {
+							substlast = g_slist_remove_link (substlast, li);
+							li = substlast->next;
+						} else {
+							subst = g_slist_remove_link (subst, subst);
+							li = subst;
+						}
+						continue;
+					}
+
 				}
 
 				substlast = li;
+				li = li->next;
 			}
 		}
 
 		for (li = dict; li != NULL; li = li->next) {
 			GelEFunc *func = li->data;
 			func->id->refs = g_slist_remove(func->id->refs,func);
-			func->id->curref =
-				func->id->refs?func->id->refs->data:NULL;
 		}
 		context.top--;
 
-		context.stack = g_slist_delete_link (context.stack, context.stack);
-		context.subststack = g_slist_delete_link (context.subststack, context.subststack);
-		context.stackname = g_slist_delete_link (context.stackname, context.stackname);
+		of = context.stack;
+		context.stack = of->next;
+
+		g_slice_free (GelContextFrame, of);
 
 		/* substitute lower variables unless we are on the toplevel */
 		if (substlast != NULL && context.top > 0) {
-			substlast->next = context.subststack->data;
-			context.subststack->data = subst;
+			substlast->next = context.stack->substlist;
+			context.stack->substlist = subst;
 			subst = NULL;
 		}
 
@@ -740,11 +880,13 @@ d_getcontext (void)
 {
 	if (context.top == -1)
 		return NULL;
-	else
-		return context.stack->data;
+	else {
+		g_assert (context.stack != NULL);
+		return context.stack->functions;
+	}
 }
 
-GSList *
+GelContextFrame *
 d_get_all_contexts (void)
 {
 	if (context.top == -1)
@@ -753,15 +895,6 @@ d_get_all_contexts (void)
 		return context.stack;
 }
 
-GSList *
-d_get_context_names (void)
-{
-	if (context.top == -1)
-		return NULL;
-	else
-		return context.stackname;
-}
-
 /*gimme the current global dictinary*/
 GSList *
 d_getcontext_global (void)
@@ -769,14 +902,23 @@ d_getcontext_global (void)
 	if (context.top == -1) {
 		return NULL;
 	} else {
-		GSList *last;
-		last = g_slist_last (context.stack);
-		g_assert (last != NULL);
-
-		return last->data;
+		g_assert (context.stack != NULL);
+		return context.global_frame->functions;
 	}
 }
 
+static int
+lowercase_ascii_sum_square (const char *id)
+{
+	int sum = 0;
+	int i;
+	for (i = 0; id[i] != '\0'; i++) {
+		int n = g_ascii_tolower (id[i]) - 'a';
+		sum += n*n;
+	}
+	return sum;
+}
+
 static int
 lowercase_ascii_sum (const char *id)
 {
@@ -850,13 +992,16 @@ are_ids_similar (const char *id1, const char *id2)
 
 	if (len1 > 6 && len1 == len2) {
 		int sum1, sum2;
+		int sum1s, sum2s;
 
 		sum1 = lowercase_ascii_sum (id1);
 		sum2 = lowercase_ascii_sum (id2);
 
-		/* just a reordering (possibly)
-		   (won't work right on small words) */
-		if (sum1 == sum2) {
+		sum1s = lowercase_ascii_sum_square (id1);
+		sum2s = lowercase_ascii_sum_square (id2);
+
+		/* just a reordering (possibly) */
+		if (sum1 == sum2 && sum1s == sum2s) {
 			return TRUE;
 		}
 	}
@@ -943,8 +1088,7 @@ d_protect_all(void)
 		GelEFunc *func = li->data;
 		if(!func->id || strcmp(func->id->token,"Ans")==0)
 			continue;
-		if ( ! func->id->parameter)
-			func->id->protected_ = 1;
+		func->id->protected_ = 1;
 	}
 }
 
