@@ -1,11 +1,13 @@
 /* GENIUS Calculator
- * Copyright (C) 1997-2006 George Lebl
+ * Copyright (C) 1997-2011 Jiri (George) Lebl
  *
  * Author: George Lebl
  *
- * This program is free software; you can redistribute it and/or modify
+ * This file is part of Genius.
+ *
+ * Genius is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -14,9 +16,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the  Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
- * USA.
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -25,6 +25,7 @@
 #include <string.h>
 #include "structs.h"
 #include "eval.h"
+#include "calc.h"
 #include "matrix.h"
 
 /*implement the inline functions*/
@@ -45,7 +46,8 @@ struct _GelMatrixWFreeList {
 	GelMatrixWFreeList *next;
 };
 
-static void internal_make_private (GelMatrixW *m, int w, int h);
+static void internal_make_private (GelMatrixW *m, int w, int h,
+				   gboolean kill_type_caches);
 
 static GelMatrixWFreeList *free_matrices = NULL;
 
@@ -73,8 +75,8 @@ internal_matrix_free (GelMatrix *m)
 #ifdef MATRIX_DEBUG
 		/*debug*/printf ("ACTUALLY FREE\n");
 #endif
-		for (i = 0; i < m->width; i++) {
-			for (j = 0; j < m->height; j++) {
+		for (j = 0; j < m->height; j++) {
+			for (i = 0; i < m->width; i++) {
 				GelETree *t = gel_matrix_index (m, i, j);
 				if (t != NULL)
 					gel_freetree (t);
@@ -142,9 +144,6 @@ gel_matrixw_new(void)
 	m->regw = m->m->width;
 	m->regh = m->m->height;
 	
-	if (the_zero == NULL)
-		the_zero = gel_makenum_ui (0);
-	
 	return m;
 }
 GelMatrixW *
@@ -159,7 +158,6 @@ gel_matrixw_new_with_matrix(GelMatrix *mat)
 	m->m = mat;
 	m->m->use++;
 
-	/* clear caches as we're gonna mess with this matrix */
 	m->cached_value_only = 0;
 	m->cached_value_only_real = 0;
 	m->cached_value_only_rational = 0;
@@ -173,8 +171,100 @@ gel_matrixw_new_with_matrix(GelMatrix *mat)
 	m->regw = m->m->width;
 	m->regh = m->m->height;
 	
-	if (the_zero == NULL)
-		the_zero = gel_makenum_ui (0);
+	return m;
+}
+
+GelMatrixW *
+gel_matrixw_new_with_matrix_value_only (GelMatrix *mat)
+{
+	GelMatrixW *m;
+	NEW_MATRIX (m);
+#ifdef MATRIX_DEBUG
+	/*debug*/printf ("%s\n", G_GNUC_PRETTY_FUNCTION);
+#endif
+	
+	m->m = mat;
+	m->m->use++;
+
+	m->cached_value_only = 1;
+	m->value_only = 1;
+	m->cached_value_only_real = 0;
+	m->cached_value_only_rational = 0;
+	m->cached_value_only_integer = 0;
+	m->cached_value_or_bool_only = 0;
+	m->rref = 0;
+	
+	m->tr = 0;
+	m->regx = NULL;
+	m->regy = NULL;
+	m->regw = m->m->width;
+	m->regh = m->m->height;
+	
+	return m;
+}
+
+GelMatrixW *
+gel_matrixw_new_with_matrix_value_only_integer (GelMatrix *mat)
+{
+	GelMatrixW *m;
+	NEW_MATRIX (m);
+#ifdef MATRIX_DEBUG
+	/*debug*/printf ("%s\n", G_GNUC_PRETTY_FUNCTION);
+#endif
+	
+	m->m = mat;
+	m->m->use++;
+
+	m->cached_value_only = 1;
+	m->value_only = 1;
+	m->cached_value_only_real = 1;
+	m->value_only_real = 1;
+	m->cached_value_only_rational = 1;
+	m->value_only_rational = 1;
+	m->cached_value_only_integer = 1;
+	m->value_only_integer = 1;
+	m->cached_value_or_bool_only = 1;
+	m->value_or_bool_only = 1;
+	m->rref = 0;
+	
+	m->tr = 0;
+	m->regx = NULL;
+	m->regy = NULL;
+	m->regw = m->m->width;
+	m->regh = m->m->height;
+	
+	return m;
+}
+
+GelMatrixW *
+gel_matrixw_new_with_matrix_value_only_real_nonrational (GelMatrix *mat)
+{
+	GelMatrixW *m;
+	NEW_MATRIX (m);
+#ifdef MATRIX_DEBUG
+	/*debug*/printf ("%s\n", G_GNUC_PRETTY_FUNCTION);
+#endif
+	
+	m->m = mat;
+	m->m->use++;
+
+	m->cached_value_only = 1;
+	m->value_only = 1;
+	m->cached_value_only_real = 1;
+	m->value_only_real = 1;
+	m->cached_value_only_rational = 1;
+	m->value_only_rational = 0;
+	m->cached_value_only_integer = 1;
+	m->value_only_integer = 0;
+	m->cached_value_or_bool_only = 1;
+	m->value_or_bool_only = 1;
+	m->rref = 0;
+	
+	m->tr = 0;
+	m->regx = NULL;
+	m->regy = NULL;
+	m->regw = m->m->width;
+	m->regh = m->m->height;
 	
 	return m;
 }
@@ -202,15 +292,15 @@ make_us_a_copy (GelMatrixW *m, int neww, int newh)
 	gel_matrix_set_size (m->m, neww, newh, TRUE /* padding */);
 	w = MIN (neww, m->regw);
 	h = MIN (newh, m->regh);
-	for (i = 0; i < w; i++) {
-		for(j = 0; j < h; j++) {
+	for(j = 0; j < h; j++) {
+		for (i = 0; i < w; i++) {
 			int mi = m->regx ? m->regx[i] : i;
 			int mj = m->regy ? m->regy[j] : j;
 			GelETree *t = gel_matrix_index (old, mi, mj);
 #ifdef MATRIX_DEBUG
 			printf ("(%d,%d) = (%d,%d)\n", i, j,mi, mj);
 #endif
-			gel_matrix_index (m->m, i, j) = copynode (t);
+			gel_matrix_index (m->m, i, j) = gel_copynode (t);
 		}
 	}
 
@@ -244,12 +334,12 @@ copy_with_region (GelMatrixW *m, int *regx, int *regy, int w, int h)
 	gel_matrix_set_size (m->m, w, h, TRUE /* padding */);
 	cw = MIN (w, m->regw);
 	ch = MIN (h, m->regh);
-	for (i = 0; i < cw; i++) {
-		for(j = 0; j < ch; j++) {
+	for(j = 0; j < ch; j++) {
+		for (i = 0; i < cw; i++) {
 			int mi = m->regx ? m->regx[regx[i]] : regx[i];
 			int mj = m->regy ? m->regy[regy[j]] : regy[j];
 			GelETree *t = gel_matrix_index (old, mi, mj);
-			gel_matrix_index (m->m, i, j) = copynode (t);
+			gel_matrix_index (m->m, i, j) = gel_copynode (t);
 		}
 	}
 
@@ -283,12 +373,15 @@ copy_internal_region (GelMatrixW *m, int w, int h)
 	gel_matrix_set_size (m->m, w, h, TRUE /* padding */);
 	cw = MIN (w, m->regw);
 	ch = MIN (h, m->regh);
-	for (i = 0; i < cw; i++) {
-		for(j = 0; j < ch; j++) {
+	for (j = 0; j < ch; j++) {
+		for (i = 0; i < cw; i++) {
 			int mi = m->regx ? m->regx[i] : i;
 			int mj = m->regy ? m->regy[j] : j;
 			GelETree *t = gel_matrix_index (old, mi, mj);
-			gel_matrix_index (m->m, i, j) = copynode (t);
+			if (t != NULL &&
+			    (t->type != GEL_VALUE_NODE ||
+			     ! mpw_exact_zero_p (t->val.value)))
+				gel_matrix_index (m->m, i, j) = gel_copynode (t);
 		}
 	}
 
@@ -335,8 +428,8 @@ ensure_at_least_size (GelMatrixW *m, int w, int h)
 		gel_matrix_set_size (m->m, w, h, TRUE /* padding */);
 		nw = MIN (w, m->regw);
 		nh = MIN (h, m->regh);
-		for (i = 0; i < nw; i++) {
-			for (j = 0; j < nh; j++) {
+		for (j = 0; j < nh; j++) {
+			for (i = 0; i < nw; i++) {
 				int mi = m->regx ? m->regx[i] : i;
 				int mj = m->regy ? m->regy[j] : j;
 				GelETree *t = gel_matrix_index (old, mi, mj);
@@ -346,8 +439,8 @@ ensure_at_least_size (GelMatrixW *m, int w, int h)
 				}
 			}
 		}
-		for (i = 0; i < old->width; i++) {
-			for (j = 0; j < old->height; j++) {
+		for (j = 0; j < old->height; j++) {
+			for (i = 0; i < old->width; i++) {
 				GelETree *t = gel_matrix_index (old,i,j);
 				if (t != NULL)
 					gel_freetree (t);
@@ -407,8 +500,8 @@ gel_matrixw_set_size (GelMatrixW *m, int nwidth, int nheight)
 				internal_matrix_free (old);
 				return;
 			}
-			for (i = width; i < m->regw; i++) {
-				for (j = 0; j < m->regh; j++) {
+			for (j = 0; j < m->regh; j++) {
+				for (i = width; i < m->regw; i++) {
 					int mi = m->regx ? m->regx[i] : i;
 					int mj = m->regy ? m->regy[j] : j;
 					if (gel_matrix_index (m->m, mi, mj) != NULL) {
@@ -417,8 +510,8 @@ gel_matrixw_set_size (GelMatrixW *m, int nwidth, int nheight)
 					}
 				}
 			}
-			for (i = 0; i < width; i++) {
-				for (j = height; j < m->regh; j++) {
+			for (j = height; j < m->regh; j++) {
+				for (i = 0; i < width; i++) {
 					int mi = m->regx ? m->regx[i] : i;
 					int mj = m->regy ? m->regy[j] : j;
 					if (gel_matrix_index (m->m, mi, mj) != NULL) {
@@ -433,7 +526,9 @@ gel_matrixw_set_size (GelMatrixW *m, int nwidth, int nheight)
 	} else if (m->m->use > 1) {
 		/* since the use is greater than 1, we WILL get a copy of
 		 * this matrix at the right size */
-		internal_make_private (m, width, height);
+		/* it may seem we could leave caches alone, but changing size
+		 * could make those values different */
+		internal_make_private (m, width, height, TRUE /* kill_type_caches */);
 		g_assert (m->m->use == 1);
 	} else /* m->m->use == 1 */{
 		ensure_at_least_size (m, width, height);
@@ -460,7 +555,7 @@ gel_matrixw_set_at_least_size (GelMatrixW *m, int width, int height)
 
 	if (width > m->regw || height > m->regh) {
 		/* FIXME: this may be a bit inefficent */
-		gel_matrixw_make_private (m);
+		gel_matrixw_make_private (m, TRUE /* kill_type_caches */);
 		make_us_a_copy (m, MAX (width, m->regw),MAX (height, m->regh));
 		ensure_at_least_size (m, width, height);
 	}
@@ -480,16 +575,56 @@ gel_matrixw_set_element (GelMatrixW *m, int x, int y, gpointer data)
 #endif
 
 	if (m->tr)
-		internal_make_private (m, MAX(m->regh, y+1), MAX(m->regw, x+1));
+		internal_make_private (m, MAX(m->regh, y+1), MAX(m->regw, x+1),
+				       TRUE /* kill_type_caches */);
 	else
-		internal_make_private (m, MAX(m->regw, x+1), MAX(m->regh, y+1));
+		internal_make_private (m, MAX(m->regw, x+1), MAX(m->regh, y+1),
+				       TRUE /* kill_type_caches */);
 	gel_matrixw_set_at_least_size (m, x+1, y+1);
-	t = gel_matrixw_set_index (m, x, y);
+	t = gel_matrixw_get_index (m, x, y);
 	if (t != NULL)
 		gel_freetree (t);
 	gel_matrixw_set_index (m, x, y) = data;
 }
 
+/*increment element*/
+int
+gel_matrixw_incr_element (GelMatrixW *m, int x, int y, mpw_ptr by)
+{
+	GelETree *t;
+
+	g_return_if_fail (m != NULL);
+	g_return_if_fail (x >= 0);
+	g_return_if_fail (y >= 0);
+#ifdef MATRIX_DEBUG
+	/*debug*/printf ("%s\n", G_GNUC_PRETTY_FUNCTION);
+#endif
+
+	if (m->tr)
+		internal_make_private (m, MAX(m->regh, y+1), MAX(m->regw, x+1),
+				       TRUE /* kill_type_caches */);
+	else
+		internal_make_private (m, MAX(m->regw, x+1), MAX(m->regh, y+1),
+				       TRUE /* kill_type_caches */);
+	gel_matrixw_set_at_least_size (m, x+1, y+1);
+	t = gel_matrixw_get_index (m, x, y);
+	if (t == NULL) {
+		if (by == NULL)
+			gel_matrixw_set_index (m, x, y) = gel_makenum_ui (1);
+		else
+			gel_matrixw_set_index (m, x, y) = gel_makenum (by);
+	} else if (t->type == GEL_VALUE_NODE) {
+		if (by == NULL)
+			mpw_add_ui (t->val.value, t->val.value, 1);
+		else
+			mpw_add (t->val.value, t->val.value, by);
+	} else {
+		gel_errorout (_("Trying to increment a nonvalue"));
+		return FALSE;
+	}
+	return TRUE;
+}
+
 /*set vector element*/
 void
 gel_matrixw_set_velement (GelMatrixW *m, int i, gpointer data)
@@ -508,7 +643,8 @@ gel_matrixw_set_velement (GelMatrixW *m, int i, gpointer data)
 			x = i % m->regh;
 			y = i / m->regh;
 		}
-		internal_make_private (m, MAX(m->regh, y+1), MAX(m->regw, x+1));
+		internal_make_private (m, MAX(m->regh, y+1), MAX(m->regw, x+1),
+				       TRUE /* kill_type_caches */);
 	} else {
 		if (m->regh == 1) {
 			x = i;
@@ -517,15 +653,107 @@ gel_matrixw_set_velement (GelMatrixW *m, int i, gpointer data)
 			x = i % m->regw;
 			y = i / m->regw;
 		}
-		internal_make_private (m, MAX(m->regw, x+1), MAX(m->regh, y+1));
+		internal_make_private (m, MAX(m->regw, x+1), MAX(m->regh, y+1),
+				       TRUE /* kill_type_caches */);
 	}
 	gel_matrixw_set_at_least_size (m, x+1, y+1);
-	t = gel_matrixw_set_index (m, x, y);
+	t = gel_matrixw_get_index (m, x, y);
 	if (t != NULL)
 		gel_freetree (t);
 	gel_matrixw_set_index (m, x, y) = data;
 }
 
+/*increment vector element*/
+int
+gel_matrixw_incr_velement (GelMatrixW *m, int i, mpw_ptr by)
+{
+	GelETree *t;
+	int x, y;
+
+	g_return_if_fail (m != NULL);
+	g_return_if_fail (i >= 0);
+
+	if (m->tr) {
+		if (m->regw == 1) {
+			x = 0;
+			y = i;
+		} else {
+			x = i % m->regh;
+			y = i / m->regh;
+		}
+		internal_make_private (m, MAX(m->regh, y+1), MAX(m->regw, x+1),
+				       TRUE /* kill_type_caches */);
+	} else {
+		if (m->regh == 1) {
+			x = i;
+			y = 0;
+		} else {
+			x = i % m->regw;
+			y = i / m->regw;
+		}
+		internal_make_private (m, MAX(m->regw, x+1), MAX(m->regh, y+1),
+				       TRUE /* kill_type_caches */);
+	}
+	gel_matrixw_set_at_least_size (m, x+1, y+1);
+	t = gel_matrixw_get_index (m, x, y);
+	if (t == NULL) {
+		if (by == NULL)
+			gel_matrixw_set_index (m, x, y) = gel_makenum_ui (1);
+		else
+			gel_matrixw_set_index (m, x, y) = gel_makenum (by);
+	} else if (t->type == GEL_VALUE_NODE) {
+		if (by == NULL)
+			mpw_add_ui (t->val.value, t->val.value, 1);
+		else
+			mpw_add (t->val.value, t->val.value, by);
+	} else {
+		gel_errorout (_("Trying to increment a nonvalue"));
+		return FALSE;
+	}
+	return TRUE;
+}
+
+int
+gel_matrixw_incr_region (GelMatrixW *m,
+			 int *destx, int *desty,
+			 int w, int h,
+			 mpw_ptr by)
+{
+	int i, j;
+
+	for (j = 0; j < h; j++) {
+		for (i = 0; i < w; i++) {
+			gel_matrixw_incr_element (m, destx[i], desty[j], by);
+		}
+	}
+}
+
+int
+gel_matrixw_incr (GelMatrixW *m, mpw_ptr by)
+{
+	int i, j, w, h;
+
+	w = gel_matrixw_width (m);
+	h = gel_matrixw_height (m);
+
+	for (j = 0; j < h; j++) {
+		for (i = 0; i < w; i++) {
+			gel_matrixw_incr_element (m, i, j, by);
+		}
+	}
+}
+
+int
+gel_matrixw_incr_vregion (GelMatrixW *m,
+			  int *desti, int len, mpw_ptr by)
+{
+	int i;
+
+	for (i = 0; i < len; i++) {
+		gel_matrixw_incr_velement (m, desti[i], by);
+	}
+}
+
 /*make sure it's in range first!*/
 GelMatrixW *
 gel_matrixw_get_region (GelMatrixW *m, int *regx, int *regy, int w, int h)
@@ -617,9 +845,9 @@ gel_matrixw_get_vregion (GelMatrixW *m, int *reg, int len)
 	gel_matrix_set_size (vec, len, 1, FALSE /* padding */);
 
 	for (i = 0; i < len; i++) {
-		GelETree *t = gel_matrixw_set_vindex (m, reg[i]);
+		GelETree *t = gel_matrixw_get_vindex (m, reg[i]);
 		if (t != NULL)
-			t = copynode (t);
+			t = gel_copynode (t);
 		gel_matrix_index (vec, i, 0) = t;
 	}
 
@@ -660,13 +888,14 @@ gel_matrixw_set_region(GelMatrixW *m, GelMatrixW *src,
 
 	/* FIXME: we will copy some nodes we don't need to copy
 	 * as we will free them just below */
-	internal_make_private (m, MAX (xmax+1, m->regw), MAX (ymax+1, m->regh));
+	internal_make_private (m, MAX (xmax+1, m->regw), MAX (ymax+1, m->regh),
+			       TRUE /* kill_type_caches */);
 	ensure_at_least_size (m, xmax+1, ymax+1);
 	/* assume that's what ensure/make_us_a_copy does */
 	g_assert (m->regx == NULL && m->regy == NULL);
 
-	for (i = 0; i < w; i++) {
-		for ( j = 0; j < h; j++) {
+	for (j = 0; j < h; j++) {
+		for (i = 0; i < w; i++) {
 			int si, sj;
 			GelETree *t = gel_matrix_index (m->m, destx[i], desty[j]);
 			if (m->tr == src->tr) {
@@ -681,7 +910,7 @@ gel_matrixw_set_region(GelMatrixW *m, GelMatrixW *src,
 			    sj >= src->regh) {
 				gel_matrix_index (m->m, destx[i], desty[j]) = NULL;
 			} else {
-				gel_matrix_index (m->m, destx[i], desty[j]) = copynode
+				gel_matrix_index (m->m, destx[i], desty[j]) = gel_copynode
 					(gel_matrix_index (src->m,
 							   src->regx ? src->regx[si] : si,
 							   src->regy ? src->regy[sj] : sj));
@@ -723,15 +952,16 @@ gel_matrixw_set_region_etree (GelMatrixW *m, GelETree *src,
 	xmax = getmax (destx, w);
 	ymax = getmax (desty, h);
 
-	internal_make_private (m, MAX (xmax+1, m->regw), MAX (ymax+1, m->regh));
+	internal_make_private (m, MAX (xmax+1, m->regw), MAX (ymax+1, m->regh),
+			       TRUE /* kill_type_caches */);
 	ensure_at_least_size (m, xmax+1, ymax+1);
 	/* assume that's what ensure/make_us_a_copy does */
 	g_assert (m->regx == NULL && m->regy == NULL);
 
-	for (i = 0; i < w; i++) {
-		for ( j = 0; j < h; j++) {
+	for (j = 0; j < h; j++) {
+		for (i = 0; i < w; i++) {
 			GelETree *t = gel_matrix_index (m->m, destx[i], desty[j]);
-			gel_matrix_index (m->m, destx[i], desty[j]) = copynode (src);
+			gel_matrix_index (m->m, destx[i], desty[j]) = gel_copynode (src);
 			if (t != NULL)
 				gel_freetree (t);
 		}
@@ -791,8 +1021,8 @@ gel_matrixw_rowsof (GelMatrixW *source)
 	for (i = 0; i < height; i++) {
 		GelETree *n;
 
-		GET_NEW_NODE (n);
-		n->type = MATRIX_NODE;
+		GEL_GET_NEW_NODE (n);
+		n->type = GEL_MATRIX_NODE;
 		n->mat.matrix = gel_matrixw_get_region (source, reg, &i, width, 1);
 
 		gel_matrix_index (mm, 0, i) = n;
@@ -829,8 +1059,8 @@ gel_matrixw_columnsof (GelMatrixW *source)
 	for (i = 0; i < width; i++) {
 		GelETree *n;
 
-		GET_NEW_NODE (n);
-		n->type = MATRIX_NODE;
+		GEL_GET_NEW_NODE (n);
+		n->type = GEL_MATRIX_NODE;
 		n->mat.matrix = gel_matrixw_get_region (source, &i, reg, 1, height);
 		n->mat.quoted = 0;
 
@@ -860,14 +1090,14 @@ gel_matrixw_diagonalof (GelMatrixW *source)
 	len = MIN(width,height);
 
 	mm = gel_matrix_new ();
-	gel_matrix_set_size (mm, len, 1, FALSE /* padding */);
+	gel_matrix_set_size (mm, 1, len, FALSE /* padding */);
 
 	for (i = 0; i < len; i++) {
-		GelETree *n = gel_matrixw_set_index (source, i, i);
+		GelETree *n = gel_matrixw_get_index (source, i, i);
 		if (n != NULL)
-			n = copynode (n);
+			n = gel_copynode (n);
 
-		gel_matrix_index (mm, i, 0) = n;
+		gel_matrix_index (mm, 0, i) = n;
 	}
 		
 	return gel_matrixw_new_with_matrix (mm);
@@ -896,18 +1126,21 @@ gel_matrixw_transpose(GelMatrixW *m)
 
 /*make private copy of the GelMatrix*/
 static void
-internal_make_private (GelMatrixW *m, int w, int h)
+internal_make_private (GelMatrixW *m, int w, int h, gboolean kill_type_caches)
 {
 #ifdef MATRIX_DEBUG
 	/*debug*/printf ("%s\n", G_GNUC_PRETTY_FUNCTION);
 #endif
 
-	/* clear caches as we're gonna mess with this matrix */
-	m->cached_value_only = 0;
-	m->cached_value_only_real = 0;
-	m->cached_value_only_rational = 0;
-	m->cached_value_only_integer = 0;
-	m->cached_value_or_bool_only = 0;
+	if (kill_type_caches) {
+		/* clear caches as we're gonna mess with this matrix */
+		m->cached_value_only = 0;
+		m->cached_value_only_real = 0;
+		m->cached_value_only_rational = 0;
+		m->cached_value_only_integer = 0;
+		m->cached_value_or_bool_only = 0;
+	}
+
 	m->rref = 0;
 
 #ifdef MATRIX_DEBUG
@@ -931,14 +1164,14 @@ internal_make_private (GelMatrixW *m, int w, int h)
 
 /*make private copy of the GelMatrix*/
 void
-gel_matrixw_make_private (GelMatrixW *m)
+gel_matrixw_make_private (GelMatrixW *m, gboolean kill_type_caches)
 {
 	g_return_if_fail(m != NULL);
 #ifdef MATRIX_DEBUG
 	/*debug*/printf ("%s\n", G_GNUC_PRETTY_FUNCTION);
 #endif
 
-	internal_make_private (m, m->regw, m->regh);
+	internal_make_private (m, m->regw, m->regh, kill_type_caches);
 }
 
 /*free a matrix*/
@@ -992,11 +1225,13 @@ gel_matrixw_set_vregion (GelMatrixW *m, GelMatrixW *src, int *desti, int len)
 	if (m->tr) {
 		int i;
 		if (m->regw == 1) {
-			internal_make_private (m, m->regw, MAX (max+1, m->regh));
+			internal_make_private (m, m->regw, MAX (max+1, m->regh),
+					       TRUE /* kill_type_caches */);
 			ensure_at_least_size (m, m->regw, max+1);
 		} else {
 			int minw = (max / m->regh) + 1;
-			internal_make_private (m, MAX (minw, m->regw), m->regh);
+			internal_make_private (m, MAX (minw, m->regw), m->regh,
+					       TRUE /* kill_type_caches */);
 			ensure_at_least_size (m, minw, m->regh);
 		}
 		/* assume that's what ensure/make_us_a_copy does */
@@ -1017,9 +1252,9 @@ gel_matrixw_set_vregion (GelMatrixW *m, GelMatrixW *src, int *desti, int len)
 			if (i >= srcelts) {
 				gel_matrix_index (m->m, x, y) = NULL;
 			} else {
-				GelETree *s = gel_matrixw_set_vindex (src, i);
+				GelETree *s = gel_matrixw_get_vindex (src, i);
 				if (s != NULL)
-					s = copynode (s);
+					s = gel_copynode (s);
 				gel_matrix_index (m->m, x, y) = s;
 			}
 
@@ -1029,11 +1264,13 @@ gel_matrixw_set_vregion (GelMatrixW *m, GelMatrixW *src, int *desti, int len)
 	} else {
 		int i;
 		if (m->regh == 1) {
-			internal_make_private (m, MAX (max+1, m->regw), m->regh);
+			internal_make_private (m, MAX (max+1, m->regw), m->regh,
+					       TRUE /* kill_type_caches */);
 			ensure_at_least_size (m, max+1, m->regh);
 		} else {
 			int minh = (max / m->regw) + 1;
-			internal_make_private (m, m->regw, MAX (minh, m->regh));
+			internal_make_private (m, m->regw, MAX (minh, m->regh),
+					       TRUE /* kill_type_caches */);
 			ensure_at_least_size (m, m->regw, minh);
 		}
 		/* assume that's what ensure/make_us_a_copy does */
@@ -1054,9 +1291,9 @@ gel_matrixw_set_vregion (GelMatrixW *m, GelMatrixW *src, int *desti, int len)
 			if (i >= srcelts) {
 				gel_matrix_index (m->m, x, y) = NULL;
 			} else {
-				GelETree *s = gel_matrixw_set_vindex (src, i);
+				GelETree *s = gel_matrixw_get_vindex (src, i);
 				if (s != NULL)
-					s = copynode (s);
+					s = gel_copynode (s);
 				gel_matrix_index (m->m, x, y) = s;
 			}
 
@@ -1083,7 +1320,8 @@ gel_matrixw_set_vregion_etree (GelMatrixW *m, GelETree *src, int *desti, int len
 		int minw = (max / m->regh) + 1;
 		int i;
 
-		internal_make_private (m, MAX (minw, m->regw), m->regh);
+		internal_make_private (m, MAX (minw, m->regw), m->regh,
+				       TRUE /* kill_type_caches */);
 		ensure_at_least_size (m, minw, m->regh);
 		/* assume that's what ensure/make_us_a_copy does */
 		g_assert (m->regx == NULL && m->regy == NULL);
@@ -1092,7 +1330,7 @@ gel_matrixw_set_vregion_etree (GelMatrixW *m, GelETree *src, int *desti, int len
 			int x = desti[i] / m->regw;
 			int y = desti[i] % m->regw;
 			GelETree *t = gel_matrix_index (m->m, x, y);
-			gel_matrix_index (m->m, x, y) = copynode (src);
+			gel_matrix_index (m->m, x, y) = gel_copynode (src);
 			if (t != NULL)
 				gel_freetree (t);
 		}
@@ -1101,7 +1339,8 @@ gel_matrixw_set_vregion_etree (GelMatrixW *m, GelETree *src, int *desti, int len
 		int minh = (max / m->regw) + 1;
 		int i;
 
-		internal_make_private (m, m->regw, MAX (minh, m->regh));
+		internal_make_private (m, m->regw, MAX (minh, m->regh),
+				       TRUE /* kill_type_caches */);
 		ensure_at_least_size (m, m->regw, minh);
 		/* assume that's what ensure/make_us_a_copy does */
 		g_assert (m->regx == NULL && m->regy == NULL);
@@ -1110,9 +1349,47 @@ gel_matrixw_set_vregion_etree (GelMatrixW *m, GelETree *src, int *desti, int len
 			int x = desti[i] % m->regw;
 			int y = desti[i] / m->regw;
 			GelETree *t = gel_matrix_index (m->m, x, y);
-			gel_matrix_index (m->m, x, y) = copynode (src);
+			gel_matrix_index (m->m, x, y) = gel_copynode (src);
 			if (t != NULL)
 				gel_freetree (t);
 		}
 	}
 }
+
+#ifndef G_CAN_INLINE
+GelETree *
+gel_matrixw_index(GelMatrixW *m, int x, int y) {
+	GelETree *t = gel_matrixw_get_index (m, x, y);
+	return t?t:the_zero;
+}
+#endif
+
+#ifndef G_CAN_INLINE
+GelETree *
+gel_matrixw_vindex(GelMatrixW *m, int i) {
+	GelETree *t;
+	int w = gel_matrixw_width(m);
+	/* Avoid dividing things */
+	if (w == 1)
+		t = gel_matrixw_get_index (m, 0, i);
+	else if (gel_matrixw_height(m) == 1)
+		t = gel_matrixw_get_index (m, i, 0);
+	else
+		t = gel_matrixw_get_index (m, i % w, i / w);
+	return t ? t : the_zero;
+}
+#endif
+
+#ifndef G_CAN_INLINE
+G_INLINE_FUNC GelETree *
+gel_matrixw_get_vindex(GelMatrixW *m, int i) {
+	int w = gel_matrixw_width(m);
+	/* Avoid dividing things */
+	if (w == 1)
+		return gel_matrixw_get_index (m, 0, i);
+	else if (gel_matrixw_height(m) == 1)
+		return gel_matrixw_get_index (m, i, 0);
+	else
+		return gel_matrixw_get_index (m, i % w, i / w);
+}
+#endif
